"""Version definition."""

try:
    # pylint: disable=unused-import
    from ._scm_version import version as __version__  # noqa: WPS433, WPS436
except ImportError:
    from ._compat import importlib_metadata_get_version  # noqa: WPS433, WPS436

    __version__ = importlib_metadata_get_version(  # noqa: WPS440
        'sphinxcontrib-towncrier',
    )



"""Cross-Python compatibility helpers."""

import sys
from typing import Iterable


if sys.version_info >= (3, 8):
    from importlib.metadata import (  # noqa: WPS433
        version as importlib_metadata_get_version,
    )
    from shlex import join as shlex_join  # noqa: WPS433
else:
    # Python 3.7 and lower:
    from shlex import quote as _shlex_quote  # noqa: WPS433

    # pylint: disable-next=line-too-long
    from importlib_metadata import (  # type: ignore[no-redef, unused-ignore]  # noqa: LN002, WPS433, WPS440  # `unused-ignore` is only needed under Python 3.6
        version as importlib_metadata_get_version,
    )

    def shlex_join(split_command: Iterable[str]) -> str:  # noqa: WPS440
        """Return a shell-escaped string from *split_command*."""
        return ' '.join(_shlex_quote(arg) for arg in split_command)


__all__ = ('importlib_metadata_get_version', 'shlex_join')  # noqa: WPS410



"""Changelog fragment discovery helpers."""


from functools import lru_cache
from pathlib import Path
from typing import Optional, Set

from sphinx.util import logging


try:
    # pylint: disable=no-name-in-module
    from towncrier.build import find_fragments  # noqa: WPS433
except ImportError:
    # pylint: disable=import-self,no-name-in-module
    from towncrier import (  # type: ignore[attr-defined] # noqa: WPS433,WPS440
        find_fragments,
    )

from ._towncrier import get_towncrier_config  # noqa: WPS436


logger = logging.getLogger(__name__)


def _resolve_spec_config(
        base: Path, spec_name: Optional[str] = None,
) -> Optional[Path]:
    return base / spec_name if spec_name is not None else None


# pylint: disable=fixme
# FIXME: consider consolidating this logic upstream in towncrier
def _find_config_file(base: Path) -> Path:
    """Find the best config file."""
    candidate_names = 'towncrier.toml', 'pyproject.toml'
    candidates = list(map(base.joinpath, candidate_names))
    extant = filter(Path.is_file, candidates)
    return next(extant, candidates[-1])


# pylint: disable=fixme
# FIXME: refactor `lookup_towncrier_fragments` to drop noqas
@lru_cache(maxsize=1, typed=True)  # noqa: WPS210
def lookup_towncrier_fragments(  # noqa: WPS210
        working_dir: Optional[str] = None,
        config_path: Optional[str] = None,
) -> Set[Path]:
    """Emit RST-formatted Towncrier changelog fragment paths."""
    project_path = Path.cwd() if working_dir is None else Path(working_dir)

    final_config_path = (
        _resolve_spec_config(project_path, config_path)
        or _find_config_file(project_path)
    )

    try:
        towncrier_config = get_towncrier_config(
            project_path,
            final_config_path,
        )
    except KeyError as key_err:
        # NOTE: The error is missing key 'towncrier' or similar
        logger.warning(
            f'Missing key {key_err!s} in file {final_config_path!s}',
        )
        return set()

    fragment_directory: Optional[str] = 'newsfragments'
    try:
        fragment_base_directory = project_path / towncrier_config['directory']
    except KeyError:
        assert fragment_directory is not None
        fragment_base_directory = project_path / fragment_directory
    else:
        fragment_directory = None

    _fragments, fragment_filenames = find_fragments(
        str(fragment_base_directory),
        towncrier_config['sections'],
        fragment_directory,
        towncrier_config['types'],
    )
    return set(fragment_filenames)



"""Towncrier related shims."""

from contextlib import suppress as _suppress_exception
from pathlib import Path
from typing import Any, Dict, Union


with _suppress_exception(ImportError):
    # NOTE: This will not raise an exception under Python >= 3.7, and is only
    # NOTE: needed for Towncrier >= 22.12.0rc1 which doesn't support Python 3.6
    from dataclasses import asdict as _dataclass_to_dict  # noqa: WPS433


try:
    # Towncrier >= 22.8.0rc1
    # pylint: disable=import-error,no-name-in-module
    from towncrier._settings.load import (  # noqa: WPS433, WPS436
        load_config_from_file,
    )
except ImportError:
    # pylint: disable=import-error,no-name-in-module
    from towncrier._settings import (  # noqa: WPS433, WPS436, WPS440
        load_config_from_file,
    )


def get_towncrier_config(
        project_path: Path,
        final_config_path: Union[Path, None],
) -> Dict[str, Any]:  # FIXME: add a better type  # pylint: disable=fixme
    """Return the towncrier config dictionary."""
    try:
        # Towncrier >= 19.9.0
        config = load_config_from_file(
            str(project_path), str(final_config_path),
        )
    except TypeError:
        # Towncrier < 19.9.0

        if final_config_path is None or not final_config_path.exists():
            # Towncrier < 19.9.0
            raise FileNotFoundError(
                f"[Errno 2] No such file or directory: '{final_config_path}'",
            ) from None

        return load_config_from_file(str(final_config_path))

    if isinstance(config, dict):
        # Towncrier < 22.12.0rc1
        return config

    return _dataclass_to_dict(config)
